# (c) Copyright 2016-2017, Synapse Wireless Inc.
"""Helper functions to control the XILINX on Mesh Board 2s"""


import time

import ftd2xxserialwin32
import ftd2xxserialutil


MESH_BOARD2_FTDI_INIT = "\x82\xe6\xe6\x82\xa6\xe6\x82\xe6\xe6"  # Set all pins as outputs and pull high
MESH_BOARD2_FTDI_RESET_LOW = "\x82\x7f\xe6"
MESH_BOARD2_FTDI_RESET_HIGH = "\x82\xe6\xe6"


def mb2_radio_select_str(module_num):
    """
    Gets the mux command string to select the desired radio module
    :param int module_num: The Mesh Board 2 radio module number
    :return: A mux command string
    :rtype: str
    """
    buf = []
    db = [
        (module_num & 0x01) << 2,
        (module_num & 0x02) << 1,
        (module_num & 0x04),
        (module_num & 0x08) >> 1
    ]
    clk_p = 0x20
    cs = 0x02
    write_mask = '\xe6'

    db_idx = 0
    for i in range(0, 24, 3):
        # Set command byte
        buf.append('\x82')
        # Set value byte
        buf.append(chr((0 if (i % 2) > 0 else clk_p) | db[db_idx] & ~cs | 0xc0))
        # Set pin mask
        buf.append(write_mask)
        if (i % 2) > 0:
            db_idx += 1

    # End transaction
    # Set command byte
    buf.append('\x82')
    # Set value byte
    buf.append(write_mask)  # everything back to high
    # Set pin mask
    buf.append(write_mask)

    return ''.join(buf)


def reset_radio(ftdi_library_path, device_num):
    """
    Resets the currently selected radio module
    :param str ftdi_library_path: Full path to the FTDI driver library files
    :param int device_num: The FTDI device number
    :return:
    """
    mux = ftd2xxserialwin32.Ftd2xxSerial(pathToLibrary=ftdi_library_path,
                                         pid=ftd2xxserialutil.MESH_BOARD2_PID,
                                         requiredDesc=ftd2xxserialutil.MESH_BOARD2_DESC + ' A',
                                         port=int(device_num))
    if not mux.configureMPSSEMode():
        raise ftd2xxserialutil.Ftd2xxSerialException("Could not configure FTDI mode")

    # TODO: Test if wait is actually needed (it is included in FTDI example code)
    time.sleep(0.05)  # Wait for all the USB stuff to complete and work

    if mux.write(MESH_BOARD2_FTDI_RESET_LOW) != len(MESH_BOARD2_FTDI_RESET_LOW):
        raise ftd2xxserialutil.Ftd2xxSerialException("Could not write reset radio low")

    # TODO: Test if wait is actually needed (it is included in FTDI example code)
    time.sleep(0.05)  # Wait for all the USB stuff to complete and work

    if mux.write(MESH_BOARD2_FTDI_RESET_HIGH) != len(MESH_BOARD2_FTDI_RESET_HIGH):
        raise ftd2xxserialutil.Ftd2xxSerialException("Could not write reset radio high")

    # TODO: Test if wait is actually needed (it is included in FTDI example code)
    time.sleep(0.05)  # Wait for all the USB stuff to complete and work

    # Revert back out of MPSEE mode
    mux.setBitMode(0, 0)

    mux.close()


def select_radio(ftdi_library_path, device_num, module_num):
    """
    Configures the Mesh Board2 mux to switch to the desired radio
    :param str ftdi_library_path: Full path to the FTDI driver library files
    :param int device_num: The FTDI device number
    :param int module_num: The Mesh Board 2 radio module number
    :return: None
    """
    mux = ftd2xxserialwin32.Ftd2xxSerial(pathToLibrary=ftdi_library_path,
                                         pid=ftd2xxserialutil.MESH_BOARD2_PID,
                                         requiredDesc=ftd2xxserialutil.MESH_BOARD2_DESC + ' A',
                                         port=int(device_num))
    if not mux.configureMPSSEMode():
        raise ftd2xxserialutil.Ftd2xxSerialException("Could not configure FTDI mode")

    # TODO: Test if wait is actually needed (it is included in FTDI example code)
    time.sleep(0.05)  # Wait for all the USB stuff to complete and work

    if mux.write(MESH_BOARD2_FTDI_INIT) != len(MESH_BOARD2_FTDI_INIT):
        raise ftd2xxserialutil.Ftd2xxSerialException("Could not write to FTDI and configure mux")

    # TODO: Test if wait is actually needed (it is included in FTDI example code)
    time.sleep(0.05)  # Wait for all the USB stuff to complete and work

    radio_sel_str = mb2_radio_select_str(int(module_num))
    if mux.write(radio_sel_str) != len(radio_sel_str):
        raise ftd2xxserialutil.Ftd2xxSerialException("Could not write to FTDI and select radio")

    # TODO: Test if wait is actually needed (it is included in FTDI example code)
    time.sleep(0.05)  # Wait for all the USB stuff to complete and work

    # Revert back out of MPSEE mode
    mux.setBitMode(0, 0)

    mux.close()
